// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import Geo.Framework.Any.*;



public class AnyFrameDescriptorImpl extends AnyImpl implements AnyFrameDescriptor {

	// instance variables
	
	protected AnyString name;
	protected AnyContext context;
	protected AnyFrameDescriptor superFrameDescriptor;
	protected AnyDictionary contents;
	protected AnyArray cache;
	protected AnyIndexCacheSlot cache1slot;
	protected AnyIndexCacheSlot cache2slot;
	protected AnyString description;


	// constructors
	
	protected AnyFrameDescriptorImpl() {
		cache = new AnyArrayImpl();
		setContents(new AnyDictionaryImpl());
	}
	
	public AnyFrameDescriptorImpl(AnyContext context, AnyFrame frame) {
		this();
		this.context = context;
		setName(frame.at("Name").asAnyString());
		Any superFrameString = frame.at("SuperFrameDescriptor");
		if ( !(superFrameString instanceof AnyNull) ) {
			setSuperFrameDescriptor(context.getSymbolTable().at(superFrameString.asAnyString()).asAnyFrameDescriptor());
		}
		generateSlotDescriptors(frame, this);
	}
	
	public AnyFrameDescriptorImpl(AnyContext anAnyContext, AnyString name) {	
		this();
		this.context = anAnyContext;
		setName(name);
		context.getSymbolTable().insert(this);
	}
	
	public AnyFrameDescriptorImpl(AnyContext anAnyContext, String name) {
		this(anAnyContext, AnySap.stringValue(name));
	}
	
	
	// accessing
	
	public AnyDictionary getContents() {
		return contents;
	}
	
	public AnyContext getContext() {
		return context;
	}
	
	public AnyString getDescription() {
		return description;
	}
	
	public AnyString getName() {
		return name;
	}
	
	public AnyFrameDescriptor getSuperFrameDescriptor() {
		return superFrameDescriptor;
	}	
	
	public int numberOfSlots() {
		return getContents().size();
	}	
	
	public int totalNumberOfSlots() {
		if ( getSuperFrameDescriptor() == null )
			return numberOfSlots();
		else
			return numberOfSlots() + getSuperFrameDescriptor().totalNumberOfSlots();
	}

	public AnyArray slotNames() {
		return getContents().keys();
	}
	
	public AnyArray allSlotNames() {
		AnyArray names;
		if ( getSuperFrameDescriptor() != null ) 
			names = getSuperFrameDescriptor().allSlotNames();
		else 
			names = new AnyArrayImpl();
		for (int i=0, s=slotNames().size(); i<s; i++)
			names.append(slotNames().at(i));
		return names;
	}	
		
		
	// setting
	
	public void setContents(AnyDictionary anAnyDictionary) {
		contents = anAnyDictionary;
		initCache();
	}
	
	public void setDescription(AnyString description) {
		this.description = description;
	}
	
	public void setDescription(String description) {
		setDescription(AnySap.stringValue(description));
	}
	
	public void setName(AnyString anAnyString) {
		name = anAnyString;
	}
	
	public void setName(String aString) {
		setName(AnySap.stringValue(aString));
	}
	
	public void setSuperFrameDescriptor(AnyFrameDescriptor anAnyFrameDescriptor) {
		superFrameDescriptor = anAnyFrameDescriptor;
		initCache();
	}	
	
	
	// accessing slots
	
	public AnySlotDescriptor at(AnyString key) {
		return (AnySlotDescriptor) getContents().at(key);
	}
	
	public AnySlotDescriptor at(String key) {
		return at(AnySap.stringValue(key));
	}

	public int offset(AnyString key) {
		if ( cache1slot != null && cache1slot.slotName().equals(key) ) 
			return cache1slot.index();
		if ( cache2slot != null && cache2slot.slotName().equals(key) )
			return cache2slot.index();
		// binary seach over cache 
		int last = cache.size()-1;
		int base = 0;
		int result, position;
		while ( last >= base ) {
			position = ( base + last ) / 2;
			AnyIndexCacheSlot currentSlot = ((AnyIndexCacheSlot) cache.at(position));
			if ( ( result = currentSlot.slotName().asString().compareTo(key.asString()) ) == 0 ) {
				cache2slot = cache1slot;
				cache1slot = currentSlot;
				return cache1slot.index();
			}
			if ( result > 0 )
				last = position - 1;
			else
				base = position + 1;
		}
		return -1;
	}	
	
	
	// manipulating slot
	
	public void append(AnyString key, AnySlotDescriptor desc) {
		if (desc == null) throw new AnyNullPointerException(this);
		getContents().atPut(key, desc);
		initCache();
	}
	
	public void append(String key, AnySlotDescriptor desc) {
		append(AnySap.stringValue(key), desc);
	}

	public AnySlotDescriptor remove(AnyString key) {
		AnySlotDescriptor removedSlotDescriptor;
		removedSlotDescriptor = (AnySlotDescriptor) getContents().remove(key);
		initCache();
		return removedSlotDescriptor;
	}
	
	public AnySlotDescriptor remove(String key) {
		return remove(AnySap.stringValue(key));
	}

	
	// comparing

	public boolean isEqual(Any anAny) {
		// equals means same name and parent
		if ( anAny == null || !(anAny instanceof AnyFrameDescriptor) )
			return false;
		AnyFrameDescriptor anAnyFrameDescriptor = anAny.asAnyFrameDescriptor();
		if ( getName().notEqual(anAnyFrameDescriptor.getName()) )
			return false;
		if ( getSuperFrameDescriptor() == null 
			&& anAnyFrameDescriptor.getSuperFrameDescriptor() != null )
				return false;

		return ( getSuperFrameDescriptor() == null && anAnyFrameDescriptor.getSuperFrameDescriptor() == null 
			||	getSuperFrameDescriptor().isEqual(anAnyFrameDescriptor.getSuperFrameDescriptor()) );
	}

	public int hashCode() {
		if ( getSuperFrameDescriptor() == null ) 
			return getName().hashCode();
		else
			return getName().hashCode() + getSuperFrameDescriptor().hashCode();
	}
	
	
	// copying

	public Any shallowCopy() {
		AnyFrameDescriptorImpl newAnyFrameDescriptor = (AnyFrameDescriptorImpl) super.shallowCopy();
		newAnyFrameDescriptor.context = context;
		newAnyFrameDescriptor.setSuperFrameDescriptor(getSuperFrameDescriptor());
		newAnyFrameDescriptor.setName(name);
		newAnyFrameDescriptor.contents = contents.shallowCopy().asAnyDictionary();
		return (Any) newAnyFrameDescriptor;
	}	

	public Any deepCopy() {
		AnyFrameDescriptorImpl newAnyFrameDescriptor =(AnyFrameDescriptorImpl) super.deepCopy();
		newAnyFrameDescriptor.setName(name);
		newAnyFrameDescriptor.context = context;
		newAnyFrameDescriptor.setSuperFrameDescriptor(getSuperFrameDescriptor());
		newAnyFrameDescriptor.contents = (AnyDictionary) contents.deepCopy();
		newAnyFrameDescriptor.initCache();
		return newAnyFrameDescriptor;
	}	
	
	
	// testing

	public boolean isSimple() {
		return false;
	}


	// printing

	public String toString() {
		return "anAnyFrameDescriptor: " + getName();
	}


	// misc
	
	public AnyFrame convertToFrame() { 
		if ( context.getSymbolTable().at(AnySap.stringValue("FrameDescriptor")) == null ) {
			AnyFrameDescriptor dummy;  
			dummy = new AnyFrameDescriptorImpl(context, "FrameDescriptor");
			dummy.append("Name", new AnySlotDescriptorImpl("String", true, false, false, false));
			dummy.append("SuperFrameDescriptor", new AnySlotDescriptorImpl("SuperFrameDescriptor", true, false, false, false));
			dummy.append("SlotDescriptor", new AnySlotDescriptorImpl("SlotDescriptor"));
			
			dummy = new AnyFrameDescriptorImpl(context, "SlotDescriptor");
			dummy.append("Name", new AnySlotDescriptorImpl("String", true, false, false, false));
			dummy.append("TypeName", new AnySlotDescriptorImpl("String", true, false, false, false));
			dummy.append("Flags", new AnySlotDescriptorImpl("String"));
		}
		AnyFrame frameDescriptorFrame = new AnyFrameImpl(context, "FrameDescriptor");
		frameDescriptorFrame.append("Name", getName());
		if ( getSuperFrameDescriptor() != null )
			frameDescriptorFrame.append("SuperFrameDescriptor", getSuperFrameDescriptor().getName());
		generateSlotInformation(frameDescriptorFrame);
		return frameDescriptorFrame;
	}	


	// private

	private void initCache() {
		cache.removeAll();
		cache1slot = null;
		cache2slot = null;
		for (int i=0, s=totalNumberOfSlots(); i<s; i++) 
			cache.append(new AnyIndexCacheSlot(allSlotNames().at(i).asAnyString(), i));
		cache.sort();
	}
	
	private AnyArray slotDescriptors() {
		return getContents().values();
	}
	
	private void generateSlotDescriptors(AnyFrame frame, AnyFrameDescriptor frameDescriptor) {
		AnyArray slots = frame.arrayAt("SlotDescriptor");
		for (int i=0, s=slots.size(); i<s; i++) {
			AnyFrame slotDescriptorFrame = slots.at(i).asAnyFrame();
			AnySlotDescriptor slotDescriptor = new AnySlotDescriptorImpl(slotDescriptorFrame.at("TypeName").asAnyString());
			AnyArray flags = slotDescriptorFrame.arrayAt("Flags");
			for (int j=0, t=flags.size(); j<t; j++) {
				if ( flags.at(j).asString().equals("SingleValue") )
					slotDescriptor.setSingleValue();
				if ( flags.at(j).asString().equals("MayNotRemove") )
					slotDescriptor.setMayNotRemove();
				if ( flags.at(j).asString().equals("MustHave") )
					slotDescriptor.setMustHave();
				if ( flags.at(j).asString().equals("NotEditable") )
					slotDescriptor.setNotEditable();
			}
			frameDescriptor.append(slotDescriptorFrame.at("Name").asAnyString(), slotDescriptor);
		}		
	}
	
	private void generateSlotInformation(AnyFrame anAnyFrame) {
		AnyArray slotNames = slotNames();  
		for (int i=0, s=slotNames.size(); i<s; i++) {
			AnyFrame slotDescriptorFrame = new AnyFrameImpl(context, "SlotDescriptor");
			AnySlotDescriptor slotDescriptor = at(slotNames.at(i).asAnyString());
			slotDescriptorFrame.append("Name", slotNames.at(i).asAnyString());
			slotDescriptorFrame.append("TypeName", slotDescriptor.getTypeName());
			if (slotDescriptor.isSingleValue())
				slotDescriptorFrame.append("Flags", AnySap.stringValue("SingleValue"));
			if (slotDescriptor.isMayNotRemove())
				slotDescriptorFrame.append("Flags", AnySap.stringValue("MayNotRemove"));	
			if (slotDescriptor.isMustHave())	
				slotDescriptorFrame.append("Flags", AnySap.stringValue("MustHave"));
			if (slotDescriptor.isNotEditable())
				slotDescriptorFrame.append("Flags", AnySap.stringValue("NotEditable"));
			anAnyFrame.append("SlotDescriptor", slotDescriptorFrame);
		}
	}

}
