// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import java.io.IOException;
import java.io.FileNotFoundException;
import Geo.Framework.Any.*;



public class AnySoupImpl extends AnyContextImpl implements AnySoup {

	// instance variables
	
	protected AnyArray data;
	protected AnyArray editData;
	protected IndexManager indexManager;
	protected IndexManager editIndexManager;
	protected IOManager ioManager;
	protected QueryProcessor queryProcessor;
	protected boolean isLocked;
	protected boolean isLockerActive;


	// class methods
	
	public static AnySoup readFromFile(AnyString metaFileName, AnyString dataFileName) throws IOException, FileNotFoundException {
		AnySoup newSoup = new AnySoupImpl();
		newSoup.loadMetaInformation(metaFileName);
		newSoup.load(dataFileName);
		return newSoup;
	}
	
	public static AnySoup readFromFile(AnyString fileName, AnySymbolTable symbolTable) throws IOException, FileNotFoundException {
		AnySoup newSoup = new AnySoupImpl(new FileIOManagerImpl(), new SimpleIndexManagerImpl(), symbolTable);
		newSoup.load(fileName);
		return newSoup;
	}
	
	public static AnySoup readFromFile(String fileName, AnySymbolTable symbolTable) throws IOException, FileNotFoundException {
	return readFromFile(AnySap.stringValue(fileName), symbolTable);
	}
	
	public static AnySoup readFromFile(String metaFileName, String dataFileName) throws IOException, FileNotFoundException {
		return readFromFile(AnySap.stringValue(metaFileName), AnySap.stringValue(dataFileName));
	}


	// constructors
	
	public AnySoupImpl() {
		this(new FileIOManagerImpl(), null, new AnySymbolTableImpl());
	}
	
	public AnySoupImpl(IndexManager indexManager) {
		this(new FileIOManagerImpl(), indexManager, new AnySymbolTableImpl());
	}
	
	public AnySoupImpl(IOManager ioManager) {
		this(ioManager, null, new AnySymbolTableImpl());
	}
	
	public AnySoupImpl(IOManager ioManager, AnySymbolTable symbolTable) {
		this(ioManager, null, symbolTable);
	}
	
	public AnySoupImpl(IOManager ioManager, IndexManager indexManager) {
		this(ioManager, indexManager, new AnySymbolTableImpl());
	}
	
	public AnySoupImpl(IOManager ioManager, IndexManager indexManager, AnySymbolTable symbolTable) {
		super(symbolTable);
		this.ioManager = ioManager;
		this.indexManager = indexManager;
		isLocked = false;
		isLockerActive = false;
		data = new AnyArrayImpl();
		editData = new AnyArrayImpl();;
		initialize();
	}
	
	
	// accessing
	
	public IndexManager getIndexManager() {
		if (isLocked && isLockerActive) 
			return editIndexManager;
		else
			return indexManager;
	}
	
	public int getNextId() {
		return ioManager.getNextId();
	}
	
	public AnyArray getData() {
		if (isLocked && isLockerActive) 
			return editData;
		else
			return data;
	}
		
	public AnyArray getAll() {
		AnyArray allData = new AnyArrayImpl();
		getData().addAllTo(allData);
		return allData;
	}


	// setting
	
	public void setQueryProcessor(QueryProcessor aQueryProcessor) {
		queryProcessor = aQueryProcessor;
	}	
	public void setData(AnyArray newData) {	
		if (isLocked && isLockerActive) 
			editData = newData;
		else
			data = newData;
	}
	
	
	// frame management
	
	public void add(AnyFrame anAnyFrame) {
		getData().append(anAnyFrame);
		if ( getIndexManager() != null )
			getIndexManager().enroll (anAnyFrame);
	}

	public void remove(AnyFrame anAnyFrame) {
		if ( getIndexManager() != null )
			getIndexManager().unroll(anAnyFrame);
		getData().remove(anAnyFrame);
	}	
	
	
	// IO 
	
	public void store(AnyString fileName) throws IOException, FileNotFoundException {
		ioManager.store(fileName);
	}

	public void store(String fileName) throws IOException, FileNotFoundException {
		store(AnySap.stringValue(fileName));
	}
	
	public void load(AnyString fileName) throws IOException, FileNotFoundException {
		ioManager.load(fileName);
	}
	
	public void load(String fileName) throws IOException, FileNotFoundException {
		load(AnySap.stringValue(fileName));
	}
	
	public void storeMetaInformation(AnyString fileName) throws IOException, FileNotFoundException {
		ioManager.storeMetaInformation(fileName);
	}
	
	public void storeMetaInformation(String fileName) throws IOException, FileNotFoundException {
		storeMetaInformation(AnySap.stringValue(fileName));
	}

	public void loadMetaInformation(AnyString fileName) throws IOException, FileNotFoundException {
		ioManager.loadMetaInformation(fileName);
	}
	
	public void loadMetaInformation(String fileName) throws IOException, FileNotFoundException {
		loadMetaInformation(AnySap.stringValue(fileName));
	}
	
	
	// locking
	
	public boolean lock() {
		if (isLocked) 
			return false;
		isLocked = true;
		isLockerActive = true;
		editIndexManager = indexManager.duplicateWithSettings();
		editData = data.deepCopy(this).asAnyArray();
		for (int i=0, s=editData.size(); i<s; i++) 
			editIndexManager.enroll(editData.at(i).asAnyFrame());
		return true;
	}
	
	public boolean unlock(boolean doCommit) {
		if (!isLocked) 
			return false;
		if (doCommit) {
			data = editData;
			indexManager = editIndexManager; 
		}
		editIndexManager = null;
		editData = null;
		isLocked = false;
		return true;
	}	

	public boolean isLocked() {
		return isLocked;
	}
	
	public void setClientInformation(boolean isLockerActive) {
		this.isLockerActive = isLockerActive;
	}
	
	
	// queries
	
	public AnyArray evaluateQuery (AnyString queryString) {
		if ( queryProcessor == null )
			throw new AnyNoQueryProcessorException();
		return queryProcessor.evaluateQuery(queryString);	
	}
	
	public AnyArray evaluateQuery (String queryString) {
		return evaluateQuery(AnySap.stringValue(queryString));
	}

	public void removeByQuery(AnyString queryString) {
		Any result = evaluateQuery(queryString);
		if ( !(result instanceof AnyArray) )
			return;
		AnyArray resultArray = result.asAnyArray();
		for (int i=0, s=resultArray.size(); i<s; i++) {
			for (int k=0, t=editData.size(); k<t; k++) 
				if ( getData().at(k) == resultArray.at(i) ) {
					getData().remove(k);
					t--;
					break;
				}
		}		
	}
	
	public void removeByQuery(String queryString) {
		removeByQuery(AnySap.stringValue(queryString));
	}


	// printing
	
	public void print() {
		String prettyString = (new PrettyAnyWriterImpl()).writeAny(getData());
		System.out.println(prettyString);
	}


	// private / protected
	
	protected void initialize() {
		ioManager.setSoup(this);
	}




}
